//RNN v1.6
//    Copyright (C) 2010  Roberto J. López-Sastre (robertoj.lopez@uah.es)
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "aggclustering.h"

unsigned int agg_clustering_rnn(vcl_deque < vnl_vector<double> > &X, vcl_vector <unsigned int> &labels, double agg_thres,vcl_vector< vcl_vector<double> > &cluster_centre )
{
  //This function computes an agglometarive clustering based on the construction
  //of RNN (Reciprocal Nearest Neighbor) pairs.
  //This is our own implementation of the algorithm described in
  //    -B. Leibe, A. Leonardis and B. Schiele. Robust object detection with Interleaved
  //    categorization and segmentation. IJCV, 2008.

  unsigned int  np=X.size(); //number of points
  unsigned int i=0, s=0, nc=0, num_labels=0, dim=0;
  int last=0;
  double sim;
  double l_agg_thres = (-1*agg_thres);
  ndist = 0;
  //Initialize labels with the correct size
  labels.clear();
  labels.resize(np);
  

  //Chain for the clusters R
  //This list is used to search nearest neighbors.
  vcl_deque <cluster> R(np);

  //each vector in a separate cluster (start point of the algorithm)
  for(i=0; i<np; i++)
  { 
    //update index of the vector
    R[i].data_index.push_back(i);
    //update centroid
    R[i].centroid=X[i];
    //update variance
    R[i].cvar=0;
       
  }

  //NN-Chain (the pair at the end of this chain is always a RNN)
  //This is a Nearest Neighbor chain
  vcl_deque <cluster> L;
  //Chain for similarities
  vcl_deque <double>  Lsim;
  //chain for the clusters C
  vcl_deque <cluster> C;

  //The algorithm starts with a random cluster of R.

  //Generate random point
  vnl_random rg;
  rg.reseed();
  unsigned int rp=rg.lrand32(np-1); //random integer [0,np-1]

  
  //Add to L
  L.push_back(R[rp]);
  last=0;

    
  
  //R\rp -> delete the cluster in R
  R.erase(R.begin()+rp);

  
  
  //Search the first NN to start the algorithm
  if(R.size()>0)
  {   
    //Search for the nearest neighbour of L[last]
    get_nn(&L[last],R,sim,s); //returns sim and s  
    //DEBUG
    //vcl_cout << "NN:R[" << s << "] - sim , " << sim << endl;

    //DEBUG
    ndist += R.size();
    
    //Add the NN chain
    last++;
    L.push_back(R[s]);//add to L
    R.erase(R.begin()+s);//delete from R
    Lsim.push_back(sim);//add to Lsim
    
    //Only two cluster at the begining?
    if(R.size()==0) 
    {
      //check the similarity
      if(Lsim[last-1] > l_agg_thres)
      {
	//Agglomerate clusters
	agglomerate_clusters(&L[last],&L[last-1]);
	C.push_back(L[last]);
      }
      else
      {
	//Save in C separately
	C.push_back(L[last-1]);
	C.push_back(L[last]);


      }

      L.clear(); //free memory      
	  
    }
  }
  else // R is empty
  {
    if(L.size() == 1)
      C.push_back(L[0]); // Only one vector, only one cluster
    
    L.clear();  //free memory    
  }


  //Start loop (until R is clear)
  while(R.size()>0)
  {
   
    //Search for the nearest neighbour of L[last]
    get_nn(&L[last],R,sim,s); //returns sim and s

    ndist += R.size();
    
   
    
    if(sim > Lsim[last-1]) //did we find a RNN?
    {      
      //No RNNs, add s to the NN chain
      last++;
      L.push_back(R[s]);//add to L
      R.erase(R.begin()+s);//delete from R
      Lsim.push_back(sim);//add sim to Lsim
      
      if(R.size()==0) //R has been emptied
      {
	//check the similarity
	if(Lsim[last-1] > l_agg_thres)
	{
	  //Agglomerate clusters
	  agglomerate_clusters(&L[last],&L[last-1]);
	  R.push_back(L[last]);//add to R the new cluster
	  L.pop_back();//erase L[last]
	  L.pop_back();//erase L[last-1]
	  last=last-2;
	 
	  //delete similarities
	  Lsim.pop_back();
	  if(last>=0)
	    Lsim.pop_back();  

	  //initialize the chain with the following nearest neighbour
	  if(last==0)
	  {
	    //Search for the nearest neighbour of L[last]
	    get_nn(&L[last],R,sim,s); //returns sim and s
	    
	    ndist += R.size();

	    //Add the NN chain
	    last++;
	    L.push_back(R[s]);//add to L
	    R.erase(R.begin()+s);//delete from R
	    Lsim.push_back(sim);//add to Lsim
  
	    if(R.size()==0) //R has been emptied?
	    {
	      //check the similarity
	      if(Lsim[last-1] > l_agg_thres)
	      {
		//Agglomerate clusters
		agglomerate_clusters(&L[last],&L[last-1]); 
		C.push_back(L[last]); //add the cluster to C
	      }
	      else
	      {
		//Save in C
		C.push_back(L[last-1]);
		C.push_back(L[last]);
	      }
	      L.clear(); // free memory
	    }	  
	  }	  
	}
	else
	{
	  //Add the clusters to C (separate clusters)
	  for(i=0; i<L.size(); i++)
	    C.push_back(L[i]);
	}
      }  
    }
    else //A RNN
    {
      if(Lsim[last-1] > l_agg_thres) //can they be agglomerated?
      {
	//Agglomerate clusters
	agglomerate_clusters(&L[last],&L[last-1]);

	R.push_back(L[last]);//add to R

	L.pop_back();//erase L[last]
	L.pop_back();//erase L[last-1]

	last=last-2;
	
	//delete similarities
	Lsim.pop_back();
	if(last>=0)
	  Lsim.pop_back();
		
	if(last==0) //initialize the chain of similarities with the next NN
	{
	  //Search for the nearest neighbour of L[last]
	  get_nn(&L[last],R,sim,s); //returns sim and s

	  ndist += R.size();


	  //Add the NN chain
	  last++;
	  L.push_back(R[s]);//add to L
	  R.erase(R.begin()+s);//delete from R
	  Lsim.push_back(sim);//add to Lsim
  
	  if(R.size()==0) //R has been emptied?
	  {
	    //check the similarity
	    if(Lsim[last-1] > l_agg_thres)
	    {
	      //Agglomerate clusters
	      agglomerate_clusters(&L[last],&L[last-1]); 
	      C.push_back(L[last]); //add the cluster to C
	    }
	    else
	    {
	      //Save in C
	      C.push_back(L[last-1]);
	      C.push_back(L[last]);
	    }
	    L.clear();	  
	  }	  
	}	
      }
      else //discard this chain
      {
	//Add the clusters to C (separate clusters)
	for(i=0; i<L.size(); i++)
	  C.push_back(L[i]);

	//to start a new chain later (see below)
	last=-1;
      }	     
    }

    //Do we need to start a new chain?
    if(last<0)
    {
      //Initialize a new chain
      last=0;
      Lsim.clear();
      L.clear();
      
      //random point
      rg.reseed();
      rp=rg.lrand32(R.size()-1); //random point

           
      //add to L
      L.push_back(R[rp]);
      //R\rp
      R.erase(R.begin()+rp);

      
      if(R.size()>0)
      {
	//Search for the nearest neighbour of L[last]
	get_nn(&L[last],R,sim,s); //returns sim and s
	
	ndist += R.size();



	//Add the NN chain
	last++;
	L.push_back(R[s]);//add to L
	R.erase(R.begin()+s);//delete from R
	Lsim.push_back(sim);//add to Lsim
      
	if(R.size()==0) //R has been emptied?
	{
	  //check the similarity
	  if(Lsim[last-1] > l_agg_thres)
	  {
	    //Agglomerate clusters
	    agglomerate_clusters(&L[last],&L[last-1]); 
	    C.push_back(L[last]);
	  }
	  else
	  {
	    //Save in C
	    C.push_back(L[last-1]);
	    C.push_back(L[last]);
	  }
	}
      }   
      else // R is empty
      {
	if(L.size() == 1)
	  C.push_back(L[0]); // Only one vector, only one cluster
	
	L.clear();  //free memory
	
      }
    }
    
  }//while loop

  
  //Chain C contains all the clusters
  nc=C.size();//number of clusters

  if(nc > 0)
    dim = C[0].centroid.size(); // take the dimension from the first element in C

  cluster_centre.clear();//delete the content
 
  vcl_vector <double> aux_centroid(dim);
  
  for(i=0; i<nc; i++)
  {
    //convert from vcl to vnl
    for(s=0;s<dim;s++)
      aux_centroid[s]=C[i].centroid[s];

    //add centroid
    cluster_centre.push_back(aux_centroid);
        
    num_labels+=C[i].data_index.size();

    for(s=0; s<C[i].data_index.size(); s++)
      labels[C[i].data_index[s]]=i;
  }

  //Were all the points asigned?.
  if(num_labels != np )
  {
    cout << "Warning: all the points were not assigned to a cluster!" << endl;
    cout << "Num. Labels = " << num_labels << " Num. Points = " << np << endl;
  }
  
  return nc;
}

void get_nn(cluster *V,vcl_deque <cluster> &R,double &sim,unsigned int &index)
{
  //Return the NN of V in the list R. Index is the position in R, and sim is the similarity.
  unsigned int n=R.size();
  
  if(n > 0)
  {
    vnl_vector <double> diff=V->centroid - R[0].centroid;
    double d;
    
    sim=-(V->cvar + R[0].cvar + (diff.squared_magnitude()));
    index=0;

    for(unsigned int i=1; i<n; i++)
    {
      diff=V->centroid-R[i].centroid;
      d=-(V->cvar + R[i].cvar + (diff.squared_magnitude()));
      if(d>sim)
      {
	 sim=d;
	 index=i;
      }
    }
  } 
  else
  {
    cout << "Warning: size of data matrix is 0. No nearest neighbour found!" << endl;
    index=0;
    sim=0;
  }   
  
}

void agglomerate_clusters(cluster *C1, cluster *C2)
{
  //Agglomerate two clusters
  //C1=C1+C2
  unsigned int m=C2->data_index.size();
  unsigned int n=C1->data_index.size();
 
  for(unsigned int i=0; i<m; i++)
    C1->data_index.push_back(C2->data_index[i]);

  //update centroid
  //c_new=((n*c_1)+m*(c_2))/(n+m)
  C1->centroid=((double(n)*C1->centroid)+(double(m)*C2->centroid))/(double(n+m));
  //update variance
  vnl_vector <double> diff=C1->centroid-C2->centroid;  
  C1->cvar=((double(n)*C1->cvar)+(double(m)*C2->cvar)+(((double(n*m))/(double(n+m)))*(diff.squared_magnitude())))/(double(n+m));
   
}


//Get options from the command line.
int get_options(int argc,char *argv[])
{
  int option_index=0;
   
  while(getopt_long(argc,argv,"",long_options,&option_index)>=0) 
  {
    if(optarg) 
    {
      switch(option_index) 
      { 
	// case values must match long_options
      case 1: //path to the data file
	data_file_path=strdup(optarg);
	break;
      case 2: //dimension of data
	dim=atoi(optarg);
	break;
      case 3: //number of vectors
	n=atoi(optarg);
	break;
      case 4: //path to out file
	out_file_path=strdup(optarg);
	break;
      case 5: //threshold
	thres=atof(optarg);
	thres=thres*thres;
	break;
     
      }
    }
    
    if(option_index==0) 
    {
      cout << "RNN - Reciprocal Nearest Neighbours Clustering" << endl;
      
      cout << "Usage: rnn [options]"<< endl
	   << "--help: Prints this information" << endl
	   << "--data: path to the data file"<<endl
	   << "--dim: dimension of data"<<endl
	   << "--n: number of vectors"<<endl
	   << "--thres: threshold for the agglomertive clustering (default 0.4)" <<endl
	   << "--out: path to the out file"<<endl;      
	         
      exit(1);
    } //Print Usage
  } //while for options


  return 1;
}


int read_data(vcl_deque < vnl_vector<double> >& data)
{
  unsigned int i=0,k=0;
  //Vector to read each descriptor
  vnl_vector <double> descriptor(dim);
  
  //Read data from file
  ifstream ffile(data_file_path);
  
  if( ffile.is_open() )
  {    
    for(k = 0; k < n; k++)
    {
      for(i = 0; i < dim; i++)
	ffile >> descriptor[i];
      data[k]=descriptor;
    }
    
           
  }
  else
  {
    cout << "Error reading data file. Please, check whether it exists" << endl;
    return -1;
  }
 
  //close file
  ffile.close();
  return 1;
}

int save_data(vcl_vector <unsigned int> &IDs,vcl_vector< vcl_vector<double> > &centres, float time, unsigned long num_distances)
{  
  unsigned int i=0,k=0;
   
  //Out file
  ofstream ffile(out_file_path);
  
  if( ffile.is_open() )
  {    
    //Write time
    ffile << time; ffile << endl;
    //Write num_distances
    ffile << num_distances; ffile << endl;

    //Write labels first
    for(k = 0; k < n; k++)
    {
      ffile << IDs[k];
      ffile << endl;     
    }
    
    //Write number of clusters
    ffile << nc << endl;
    
    //Write centers
    for( i=0; i < nc; i++)
    {
      for(k=0; k<dim; k++)
	ffile << centres[i][k] << " ";

      ffile << endl;
      
    }          
  }
  else
  {
    cout << "Error writing results file." << endl;
    return -1;
  }
 
  //close file
  ffile.close();
  return 1;
}


////////////////////////////////
//           MAIN
///////////////////////////////

int main(int argc, char *argv[])
{
  float t_rnn = 0;

  get_options(argc,argv);

  vcl_cout << endl;
  vcl_cout << "================" << endl;
  vcl_cout << "      RNN       " << endl;
  vcl_cout << "================" << endl << endl;
  
  vcl_cout << "Reading data ..." << endl << endl;
   
  //list with all the vectors
  vcl_deque < vnl_vector<double> > matrix_data(n);
 
  //vector for centres
  vcl_vector< vcl_vector<double> > cluster_centre;

  //vector for labels
  vcl_vector <unsigned int> labels_data(n);

  //read data from file
  read_data(matrix_data);
  
  vcl_cout << "Doing clustering. It may take a while." <<endl<<endl;
     
  //rnn clustering
  TIME_THIS(nc=agg_clustering_rnn(matrix_data, labels_data, thres,cluster_centre),t_rnn);
  vcl_cout << endl << "tRNN = " << t_rnn << endl;
  
  vcl_cout << "Clustering done!" << endl<<endl;
 
  vcl_cout << "Saving data ..." << endl<<endl;

  save_data(labels_data,cluster_centre,t_rnn,ndist);
    
  //Free memory, we will not need it anymore!
  labels_data.clear();
  matrix_data.clear();
  cluster_centre.clear();

  cout << "ThE eNd" << endl;
  
  return 0;
}
